# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

from metacar import __version__

project = "Metacar"
copyright = "2025, YDL-Simulation"
author = "YDL-Simulation"
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinxcontrib.autodoc_pydantic",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

language = "zh-CN"

autodoc_pydantic_model_show_json = False
autodoc_pydantic_model_show_field_summary = False
autodoc_pydantic_field_show_alias = False

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



__version__ = "0.3.0"

from .sceneapi import SceneAPI
from .geometry import Vector2, Vector3
from .models import (
    VLAExtension,
    VLATextOutput,
    VLAExtensionOutput,
    SubSceneInfo,
    LineType,
    BorderInfo,
    LaneInfo,
    DrivingType,
    TrafficSignType,
    RoadInfo,
    SceneStaticData,
    PoseGnss,
    GearMode,
    MainVehicleInfo,
    CameraInfo,
    SensorInfo,
    ObstacleType,
    ObstacleInfo,
    TrafficLightState,
    TrafficLightInfo,
    TrafficLightGroupInfo,
    SceneStatus,
    SimCarMsg,
    VehicleControl,
    CameraFrame,
)

__all__ = [
    "__version__",
    # sceneapi
    "SceneAPI",
    # geometry
    "Vector2",
    "Vector3",
    # models
    "VLAExtension",
    "VLATextOutput",
    "VLAExtensionOutput",
    "SubSceneInfo",
    "LineType",
    "BorderInfo",
    "LaneInfo",
    "DrivingType",
    "TrafficSignType",
    "RoadInfo",
    "SceneStaticData",
    "PoseGnss",
    "GearMode",
    "MainVehicleInfo",
    "CameraInfo",
    "SensorInfo",
    "ObstacleType",
    "ObstacleInfo",
    "TrafficLightState",
    "TrafficLightInfo",
    "TrafficLightGroupInfo",
    "SceneStatus",
    "SimCarMsg",
    "VehicleControl",
    "CameraFrame",
]



"""
使用 tkinter 和多进程实现的车辆信息显示界面。
之前也尝试过使用多线程，但是 tkinter 的组件只能在创建它的线程中使用，所以只能使用多进程。
"""

from tkinter import *
from tkinter import ttk
from multiprocessing import Process, Pipe
from multiprocessing.connection import Connection
import time
from enum import Enum, auto
import math
from dataclasses import dataclass
import numpy as np
from scipy.spatial import KDTree
from metacar import (
    Vector2,
    SceneStaticData,
    SimCarMsg,
    LineType,
)


@dataclass
class LineInfo:
    """绘制用的车道线信息"""

    type: LineType
    points: list[Vector2]


class MsgType(Enum):
    UPDATE = auto()
    QUIT = auto()


def tk_process_func(
    conn: Connection, scene_static_data: SceneStaticData, refresh_interval: int
):
    ROADLINE_COLOR_MAP = {
        LineType.MIDDLE_LINE: "dark orange",  # 中线
        LineType.SIDE_LINE: "black",  # 边线
        LineType.SOLID_LINE: "black",  # 实线
        LineType.STOP_LINE: "dark red",  # 停车线
        LineType.ZEBRA_CROSSING: "dark blue",  # 斑马线
        LineType.DASH_LINE: "grey",  # 虚线
    }

    def update(msg: SimCarMsg):
        string_vars[0].set("{:.3f}".format(msg.main_vehicle.throttle))
        string_vars[1].set("{:.3f}".format(msg.main_vehicle.brake))
        string_vars[2].set("{:.3f}".format(msg.main_vehicle.steering))
        string_vars[3].set(msg.main_vehicle.gear.name)
        string_vars[4].set("{:.3f}".format(msg.main_vehicle.speed))
        string_vars[5].set("{:.2f}".format(msg.pose_gnss.pos_x))
        string_vars[6].set("{:.2f}".format(msg.pose_gnss.pos_y))
        string_vars[7].set("{:.2f}".format(msg.pose_gnss.ori_z))
        # clear canvas
        map_canvas.delete("all")
        canvas_width = map_canvas.winfo_width()
        canvas_height = map_canvas.winfo_height()
        main_vehicle_pos = Vector2(msg.pose_gnss.pos_x, msg.pose_gnss.pos_y)

        SCALE = 10

        def convert_pos(pos: Vector2) -> tuple[float, float]:
            """将场景中的坐标转换为画布上的坐标。"""
            vec = (pos - main_vehicle_pos) * SCALE
            return (canvas_width / 2 + vec.x, canvas_height / 2 - vec.y)

        def draw_rectangle(
            center_pos: Vector2,
            length: float,
            width: float,
            ori_z: float,
            outline_color: str,
        ):
            """绘制带有旋转角度的矩形。"""
            rotate_rad = -math.radians(ori_z)
            a = Vector2(length / 2, width / 2)
            b = Vector2(length / 2, -width / 2)
            c = Vector2(-length / 2, -width / 2)
            d = Vector2(-length / 2, width / 2)
            a, b, c, d = map(
                lambda v: convert_pos(v.rotate_rad(rotate_rad) + center_pos),
                (a, b, c, d),
            )
            # 绘制无填充的多边形
            map_canvas.create_polygon(a, b, c, d, fill="", outline=outline_color)

        # 绘制主车
        draw_rectangle(
            main_vehicle_pos,
            msg.main_vehicle.length,
            msg.main_vehicle.width,
            msg.pose_gnss.ori_z,
            "green",
        )

        # 绘制障碍物
        for obstacle in msg.obstacles:
            draw_rectangle(
                Vector2(obstacle.pos_x, obstacle.pos_y),
                obstacle.length,
                obstacle.width,
                obstacle.ori_z,
                "red",
            )

        # 绘制行驶路线
        trajectory_points = [convert_pos(pt.to_vector2()) for pt in msg.trajectory]
        if len(trajectory_points) > 1:
            map_canvas.create_line(trajectory_points, fill="blue")

        # 绘制车道线
        radius = math.hypot(canvas_width, canvas_height) / 2 / SCALE
        min_idx = [-1] * len(line_list)
        max_idx = [-1] * len(line_list)
        point_idxs = kdtree.query_ball_point(
            (main_vehicle_pos.x, main_vehicle_pos.y), radius
        )
        for point_idx in point_idxs:
            rid, pid = point_info[point_idx]
            if min_idx[rid] == -1:
                min_idx[rid] = pid
                max_idx[rid] = pid
            else:
                min_idx[rid] = min(min_idx[rid], pid)
                max_idx[rid] = max(max_idx[rid], pid)
        for line_idx, line in enumerate(line_list):
            cur_min_idx = min_idx[line_idx]
            cur_max_idx = max_idx[line_idx]
            if cur_min_idx == -1:
                continue
            if line.type == LineType.ZEBRA_CROSSING:
                points = [convert_pos(Vector2(pt.x, pt.y)) for pt in line.points]
                map_canvas.create_polygon(
                    points, fill="", outline=ROADLINE_COLOR_MAP[line.type]
                )
                continue
            if cur_min_idx > 0:
                cur_min_idx -= 1
            if cur_max_idx < len(line.points) - 1:
                cur_max_idx += 1
            points = [
                convert_pos(Vector2(pt.x, pt.y))
                for pt in line.points[cur_min_idx : cur_max_idx + 1]
            ]
            map_canvas.create_line(points, fill=ROADLINE_COLOR_MAP[line.type])

        # 判断是否是 VLA 场景
        if scene_static_data.vla_extension:
            # 绘制建筑物
            for building in scene_static_data.vla_extension.buildings:
                # 绘制建筑物边框
                pos = Vector2(building.pos_x, building.pos_y)
                draw_rectangle(
                    pos,
                    building.length,
                    building.width,
                    building.ori_z,
                    "blue",
                )
                # 在建筑物中心显示名称
                center_pos = convert_pos(pos)
                map_canvas.create_text(
                    center_pos[0],
                    center_pos[1],
                    text=building.name,
                    fill="blue",
                )

    def check_message():
        """检查管道，如果有消息则处理。"""
        start_time = time.time()

        if conn.poll():
            message = conn.recv()
            match message["type"]:
                case MsgType.UPDATE:
                    update(message["msg"])
                case MsgType.QUIT:
                    root.destroy()

        # 计算下一次调用 check_message 的时间
        elapsed_time = time.time() - start_time
        next_interval = max(0, refresh_interval - int(elapsed_time * 1000))
        root.after(next_interval, check_message)

    # 先把所有需要绘制的车道线提取出来
    line_list: list[LineInfo] = []
    for road in scene_static_data.roads:
        if road.stop_line:
            line_list.append(LineInfo(LineType.STOP_LINE, road.stop_line))
        for lane in road.lanes:
            line_list.append(
                LineInfo(lane.left_border.type, lane.left_border.path_points)
            )
            line_list.append(
                LineInfo(lane.right_border.type, lane.right_border.path_points)
            )

    # 建 kdtree
    tmp_line_points: list[tuple[float, float]] = []
    point_info: list[tuple[int, int]] = []
    for line_idx, line in enumerate(line_list):
        for point_idx, point in enumerate(line.points):
            tmp_line_points.append((point.x, point.y))
            point_info.append((line_idx, point_idx))
    # 将 tmp_points 转换为 numpy 数组，并 reshape 为 (N, 2) 的形状，避免数组为空时建树失败
    line_points = np.array(tmp_line_points, dtype=float).reshape(-1, 2)
    del tmp_line_points
    kdtree = KDTree(line_points)

    root = Tk()
    root.title("仪表盘")

    mainframe = ttk.Frame(root, padding="4 4 4 4")
    mainframe.grid(column=0, row=0, sticky="NSEW")
    root.columnconfigure(0, weight=1)
    root.rowconfigure(0, weight=1)
    # 设置数值列的最小宽度，使得数值列的宽度不会随着数值的变化而变化
    mainframe.grid_columnconfigure(1, minsize=60)

    # 仪表盘的各个组件
    labels = ["油门", "刹车", "方向盘", "档位", "速度", "x坐标", "y坐标", "航向角"]
    string_vars = [StringVar() for _ in range(len(labels))]

    for row, (label_str, string_var) in enumerate(zip(labels, string_vars)):
        ttk.Label(mainframe, text=label_str).grid(column=0, row=row, sticky=E)
        ttk.Label(mainframe, textvariable=string_var).grid(column=1, row=row, sticky=W)

    # 在这里设置画布的默认大小
    map_canvas = Canvas(
        mainframe,
        width=480,
        height=480,
        borderwidth=2,
        relief="groove",
        background="white",
    )
    map_canvas.grid(column=2, row=0, rowspan=len(labels), sticky="NSEW")

    # 窗口大小改变时，将空间分配给画布
    mainframe.grid_columnconfigure(2, weight=1)
    for row in range(len(labels)):
        mainframe.grid_rowconfigure(row, weight=1)

    # 将仪表盘窗口置顶
    root.wm_attributes("-topmost", True)

    root.after_idle(check_message)
    conn.send("ready")
    root.mainloop()


class Dashboard:
    """
    显示车辆信息的仪表盘。
    使用多进程实现，使用 multiprocessing.Pipe 作为进程间通信的方式。
    """

    def __init__(self, scene_static_data: SceneStaticData, refresh_interval: int = 20):
        """
        初始化仪表盘。
        :param scene_static_data: 场景静态数据，用于绘制车道线。
        :param refresh_interval: 刷新间隔，单位毫秒，一定要比 update 方法的调用频率高。
        """
        self._conn, child_conn = Pipe()
        self._tk_process = Process(
            target=tk_process_func,
            args=(child_conn, scene_static_data, refresh_interval),
        )
        self._tk_process.start()
        # 等待仪表盘初始化完成（包括创建窗口、创建 kdtree 等）
        ready = self._conn.recv()
        if ready != "ready":
            raise RuntimeError("仪表盘初始化失败")

    def update(self, msg: SimCarMsg):
        """更新仪表盘的方法。"""
        if self._tk_process.is_alive():
            self._conn.send({"type": MsgType.UPDATE, "msg": msg})

    def quit(self):
        """退出仪表盘的方法。"""
        if self._tk_process.is_alive():
            self._conn.send({"type": MsgType.QUIT})
        self._tk_process.join()



from pydantic.dataclasses import dataclass
import math


@dataclass
class Vector2:
    x: float  #: x 坐标
    y: float  #: y 坐标

    def __iter__(self):
        return iter((self.x, self.y))

    def __pos__(self):
        return self

    def __neg__(self):
        return Vector2(*(-a for a in self))

    def __add__(self, other: "Vector2"):
        return Vector2(*(a + b for a, b in zip(self, other)))

    def __sub__(self, other: "Vector2"):
        return Vector2(*(a - b for a, b in zip(self, other)))

    def __mul__(self, other: float):
        """标量乘法。"""
        return Vector2(*(a * other for a in self))

    def __rmul__(self, other: float):
        """标量乘法。"""
        return self * other

    def __truediv__(self, other: float):
        """标量除法。"""
        return Vector2(*(a / other for a in self))

    def rotate_rad(self, radians: float) -> "Vector2":
        """绕原点旋转向量。

        将向量绕原点按逆时针方向旋转指定的弧度。

        :param radians: 旋转的角度，单位为弧度
        :type radians: float
        :return: 旋转后的新向量
        :rtype: Vector2
        """
        x = self.x * math.cos(radians) - self.y * math.sin(radians)
        y = self.x * math.sin(radians) + self.y * math.cos(radians)
        return Vector2(x, y)

    def angle_rad(self) -> float:
        """计算向量与 x 轴的夹角。

        计算二维向量与 x 轴正方向的夹角，返回值范围为 [-π, π]。

        :return: 向量与 x 轴的夹角
        :rtype: float
        """
        return math.atan2(self.y, self.x)


@dataclass
class Vector3:
    x: float  #: x 坐标
    y: float  #: y 坐标
    z: float  #: z 坐标

    def __iter__(self):
        return iter((self.x, self.y, self.z))

    def __pos__(self):
        return self

    def __neg__(self):
        return Vector3(*(-a for a in self))

    def __add__(self, other: "Vector3"):
        return Vector3(*(a + b for a, b in zip(self, other)))

    def __sub__(self, other: "Vector3"):
        return Vector3(*(a - b for a, b in zip(self, other)))

    def __mul__(self, other: float):
        """标量乘法。"""
        return Vector3(*(a * other for a in self))

    def __rmul__(self, other: float):
        """标量乘法。"""
        return self * other

    def __truediv__(self, other: float):
        """标量除法。"""
        return Vector3(*(a / other for a in self))

    def yaw_rad(self) -> float:
        """计算向量在 xOy 平面上的投影与 x 轴的夹角。

        :returns: 向量在 xOy 平面上的投影与 x 轴正方向的夹角，范围为 [-π, π]，单位为弧度
        :rtype: float
        """
        return math.atan2(self.y, self.x)

    def to_vector2(self) -> Vector2:
        """将三维向量转换为二维向量。

        该方法提取向量的 x 和 y 分量，忽略 z 分量，创建并返回一个新的二维向量。

        :returns: 包含原向量 x 和 y 分量的二维向量
        :rtype: Vector2
        """
        return Vector2(self.x, self.y)



from typing import Literal
from pydantic import BaseModel, Field
from enum import Enum
import numpy as np
from dataclasses import dataclass
from .geometry import Vector2, Vector3


class BuildingInfo(BaseModel):
    """建筑物信息"""

    id: str = Field(description="建筑物 ID")
    name: str = Field(alias="displayName", description="建筑物名称")
    pos_x: float = Field(alias="posX", description="位置 X")
    pos_y: float = Field(alias="posY", description="位置 Y")
    pos_z: float = Field(alias="posZ", description="位置 Z")
    ori_x: float = Field(alias="oriX", description="欧拉角 X（单位：角度）")
    ori_y: float = Field(alias="oriY", description="欧拉角 Y（单位：角度）")
    ori_z: float = Field(alias="oriZ", description="欧拉角 Z（单位：角度）")
    length: float = Field(description="长度")
    width: float = Field(description="宽度")
    height: float = Field(description="高度")


class VLAExtension(BaseModel):
    """VLA 扩展信息"""

    buildings: list[BuildingInfo] = Field(
        alias="BuildingInfos", description="建筑物信息"
    )


class VLATextOutput(BaseModel):
    """VLA 场景的文本输出"""

    ocr_text: str = Field(serialization_alias="OcrText", description="OCR 文本")
    time_phrase: str = Field(serialization_alias="TimeText", description="时间相关片段")
    location_phrase: str = Field(
        serialization_alias="LocationText", description="位置相关片段"
    )
    action_phrase: str = Field(
        serialization_alias="ActionText", description="动作相关片段"
    )


class VLAExtensionOutput(BaseModel):
    """VLA 场景的扩展输出"""

    text_info: VLATextOutput = Field(
        serialization_alias="TextInfo", description="文本相关输出"
    )


class SubSceneInfo(BaseModel):
    """子场景信息"""

    name: str = Field(alias="SubSceneName", description="子场景名称")
    start_point: Vector3 | None = Field(
        alias="StartPoint", description="起点（VLA 场景为 None）"
    )
    end_point: Vector3 | None = Field(
        alias="EndPoint", description="终点（VLA 场景为 None）"
    )


class MapConfig(BaseModel):
    """地图配置"""

    path: str = Field(description="地图目录路径")
    route: str = Field(description="路线文件名")
    map: str = Field(description="地图文件名")
    sub_scenes: list[SubSceneInfo] = Field(
        alias="SubSceneInfo", description="子场景信息"
    )


class LineType(Enum):
    """道路线类型"""

    MIDDLE_LINE = 1  #: 中线
    SIDE_LINE = 2  #: 侧线
    SOLID_LINE = 3  #: 实线
    STOP_LINE = 4  #: 停止线
    ZEBRA_CROSSING = 5  #: 斑马线
    DASH_LINE = 6  #: 虚线


class BorderInfo(BaseModel):
    """车道边界信息"""

    type: LineType = Field(alias="borderType", description="边界类型")
    path_points: list[Vector2] = Field(
        alias="pathPoint", description="组成边界线的点，相邻点间隔约 3~5 米"
    )


class LaneInfo(BaseModel):
    """车道信息"""

    id: str = Field(description="车道 ID")
    left_border: BorderInfo = Field(alias="LeftBorder", description="左侧边界")
    right_border: BorderInfo = Field(alias="RightBorder", description="右侧边界")
    left_lane_id: str = Field(alias="leftLane", description="左侧车道 ID")
    right_lane_id: str = Field(alias="rightLane", description="右侧车道 ID")
    width: float = Field(description="车道宽度")
    path_points: list[Vector2] = Field(alias="pathPoint", description="车道中心线")


class DrivingType(Enum):
    """行驶类型"""

    MOTOR_VEHICLE_ALLOWED = 1  #: 机动车可行驶
    NON_MOTOR_VEHICLE_ALLOWED = 2  #: 非机动车可行驶
    PEDESTRIAN_ALLOWED = 3  #: 行人可行


class TrafficSignType(Enum):
    """交通标志"""

    NO_SIGN = 0  #: 无标志
    SPEED_LIMIT_SIGN = 1  #: 限速标志
    STOP_SIGN = 2  #: 停止标志
    V2X_SIGN = 3  #: V2X 标志


class RoadInfo(BaseModel):
    """道路信息，一条道路(Road)由一个或多个车道(Lane)组成"""

    id: str = Field(description="道路 ID")
    begin_pos: Vector3 = Field(alias="beginPos", description="起点")
    end_pos: Vector3 = Field(alias="endPos", description="终点")
    driving_type: DrivingType = Field(alias="drivingType", description="行驶类型")
    traffic_sign_type: TrafficSignType = Field(
        alias="trafficSign", description="交通标志"
    )
    stop_line: list[Vector2] = Field(alias="stopLine", description="停止线位置")
    predecessor_ids: list[str] = Field(alias="predecessor", description="前驱道路 ID")
    successor_ids: list[str] = Field(alias="successor", description="后继道路 ID")
    lanes: list[LaneInfo] = Field(alias="laneData", description="车道信息")


class SceneStaticData(BaseModel):
    """场景静态信息"""

    route: list[Vector3] = Field(description="路线（VLA 场景为空列表）")
    roads: list[RoadInfo] = Field(description="道路信息（VLA 场景为空列表）")
    sub_scenes: list[SubSceneInfo] = Field(description="子场景信息")
    vla_extension: VLAExtension | None = Field(
        description="VLA 扩展信息，该字段不为 None 时表示是 VLA 场景",
    )


class PoseGnss(BaseModel):
    """车辆位姿信息"""

    pos_x: float = Field(alias="posX", description="位置 X")
    pos_y: float = Field(alias="posY", description="位置 Y")
    pos_z: float = Field(alias="posZ", description="位置 Z")
    vel_x: float = Field(alias="velX", description="速度 X")
    vel_y: float = Field(alias="velY", description="速度 Y")
    vel_z: float = Field(alias="velZ", description="速度 Z")
    ori_x: float = Field(alias="oriX", description="欧拉角 X（单位：角度）")
    ori_y: float = Field(alias="oriY", description="欧拉角 Y（单位：角度）")
    ori_z: float = Field(alias="oriZ", description="欧拉角 Z（单位：角度）")


class GearMode(Enum):
    """档位模式"""

    NEUTRAL = 0  #: 空档
    DRIVE = 1  #: 前进档
    REVERSE = 2  #: 倒车档
    PARKING = 3  #: 停车档


class MainVehicleInfo(BaseModel):
    """主车信息"""

    id: int = Field(alias="mainVehicleId", description="主车 ID")
    speed: float = Field(description="车速")
    gear: GearMode = Field(description="档位")
    throttle: float = Field(description="油门")
    brake: float = Field(description="刹车")
    steering: float = Field(description="方向盘")
    length: float = Field(description="长度")
    width: float = Field(description="宽度")
    height: float = Field(description="高度")
    left_blinker_on: bool = Field(
        alias="Signal_Light_LeftBlinker", description="左转向灯"
    )
    right_blinker_on: bool = Field(
        alias="Signal_Light_RightBlinker", description="右转向灯"
    )
    hazard_lights_on: bool = Field(alias="Signal_Light_DoubleFlash", description="双闪")
    brake_lights_on: bool = Field(alias="Signal_Light_BrakeLight", description="刹车灯")
    headlights_on: bool = Field(alias="Signal_Light_FrontLight", description="前灯")


class EulerAngle(BaseModel):
    """欧拉角"""

    # 场景传来的字段严谨一点应该为 oriX, oriY, oriZ，但实际是 orix, oriy, oriz
    # 这里先兼容一下，否则需要修改场景代码
    ori_x: float = Field(alias="orix", description="欧拉角 X（单位：角度）")
    ori_y: float = Field(alias="oriy", description="欧拉角 Y（单位：角度）")
    ori_z: float = Field(alias="oriz", description="欧拉角 Z（单位：角度）")


class CameraInfo(BaseModel):
    """摄像头信息"""

    id: str = Field(alias="Id", description="摄像头 ID")
    position: Vector3 = Field(alias="Position", description="位置")
    orientation: EulerAngle = Field(alias="Angle", description="角度")
    fov: float = Field(alias="Fov", description="视场角")
    intrinsic_matrix: list[float] = Field(
        alias="IntrinsicMatrix", description="内参矩阵"
    )
    image_width: int = Field(alias="ImageW", description="图像宽度")
    image_height: int = Field(alias="ImageH", description="图像高度")


class SensorInfo(BaseModel):
    """传感器信息"""

    ego_rgb_cams: list[CameraInfo] = Field(alias="egoRGBCams", description="主车摄像头")
    v2x_cams: list[CameraInfo] = Field(alias="v2xCams", description="V2X 摄像头")


class ObstacleType(Enum):
    """障碍物类型"""

    UNKNOWN = 0  #: 未知障碍物
    PEDESTRIAN = 4  #: 行人
    CAR = 6  #: 小汽车
    STATIC = 7  #: 静态障碍物
    BICYCLE = 8  #: 自行车
    ROAD_MARK = 12  #: 道路标记
    TRAFFIC_SIGN = 13  #: 交通标志
    TRAFFIC_LIGHT = 15  #: 交通信号灯
    RIDER = 17  #: 骑手
    TRUCK = 18  #: 卡车
    BUS = 19  #: 公交车
    SPECIAL_VEHICLE = 20  #: 特种车辆
    MOTORCYCLE = 21  #: 摩托车
    DYNAMIC = 22  #: 动态障碍物
    SPEED_LIMIT_SIGN = 26  #: 限速标志（限速值以 "SpeedLimit|30"(单位：km/h) 的格式在 :attr:`ObstacleInfo.extra_info` 中给出）
    BICYCLE_STATIC = 27  #: 静止自行车
    ROAD_OBSTACLE = 29  #: 道路障碍物
    PARKING_SLOT = 30  #: 停车位


class ObstacleInfo(BaseModel):
    """障碍物信息"""

    id: int = Field(description="障碍物 ID")
    type: ObstacleType = Field(description="障碍物类型")
    pos_x: float = Field(alias="posX", description="位置 X")
    pos_y: float = Field(alias="posY", description="位置 Y")
    pos_z: float = Field(alias="posZ", description="位置 Z")
    vel_x: float = Field(alias="velX", description="速度 X")
    vel_y: float = Field(alias="velY", description="速度 Y")
    vel_z: float = Field(alias="velZ", description="速度 Z")
    ori_x: float = Field(alias="oriX", description="欧拉角 X（单位：角度）")
    ori_y: float = Field(alias="oriY", description="欧拉角 Y（单位：角度）")
    ori_z: float = Field(alias="oriZ", description="欧拉角 Z（单位：角度）")
    length: float = Field(description="长度")
    width: float = Field(description="宽度")
    height: float = Field(description="高度")
    extra_info: str | None = Field(alias="RedundantValue", description="额外信息")


class TrafficLightState(Enum):
    """交通灯状态"""

    RED = 1  #: 红灯
    GREEN = 2  #: 绿灯
    YELLOW = 3  #: 黄灯


class TrafficLightInfo(BaseModel):
    """一排交通灯的信息"""

    id: str = Field(description="交通灯 ID")
    road_id: str = Field(alias="roadId", description="道路 ID")
    position: Vector3 = Field(alias="Position", description="位置")
    left_state: TrafficLightState = Field(alias="turnLeftState", description="左转状态")
    left_remaining_time: float = Field(
        alias="turnLeftRemainder", description="左转剩余时间"
    )
    right_state: TrafficLightState = Field(
        alias="turnRightState", description="右转状态"
    )
    right_remaining_time: float = Field(
        alias="turnRightRemainder", description="右转剩余时间"
    )
    straight_state: TrafficLightState = Field(
        alias="straightState", description="直行状态"
    )
    straight_remaining_time: float = Field(
        alias="straightRemainder", description="直行剩余时间"
    )


class TrafficLightGroupInfo(BaseModel):
    """交通灯组信息，一组交通灯共同表示一个路口的信号灯信息。"""

    id: str = Field(description="交通灯组 ID")
    traffic_lights: list[TrafficLightInfo] = Field(
        alias="trafficLightState", description="交通灯信息"
    )


class SceneStatus(BaseModel):
    """场景状态信息"""

    sub_scene_name: str = Field(alias="SubSceneName", description="子场景名称")
    used_time: float = Field(alias="UsedTime", description="已用时间")
    time_limit: float = Field(alias="TimeLimit", description="时间限制")
    end_point: Vector3 | None = Field(
        alias="EndPoint", description="终点（VLA 场景为 None）"
    )


class SimCarMsg(BaseModel):
    """仿真动态信息"""

    trajectory: list[Vector3] = Field(alias="Trajectory", description="推荐轨迹")
    pose_gnss: PoseGnss = Field(alias="PoseGnss", description="GNSS 数据")
    main_vehicle: MainVehicleInfo = Field(
        alias="DataMainVehicle", description="主车信息"
    )
    sensor: SensorInfo = Field(alias="Sensor", description="传感器信息")
    obstacles: list[ObstacleInfo] = Field(
        alias="ObstacleEntryList", description="障碍物信息"
    )
    traffic_light_groups: list[TrafficLightGroupInfo] = Field(
        alias="TrafficLightStateLists", description="交通灯组信息"
    )
    scene_status: SceneStatus = Field(alias="SceneStatus", description="场景状态信息")


class VehicleControl(BaseModel):
    """车辆控制信息"""

    throttle: float = Field(default=0.0, description="油门（0~1）")
    brake: float = Field(default=0.0, description="刹车（0~1）")
    steering: float = Field(default=0.0, description="方向盘（-1~1）")
    gear: GearMode = Field(default=GearMode.DRIVE, description="档位")
    left_blinker_on: bool = Field(default=False, description="左转向灯")
    right_blinker_on: bool = Field(default=False, description="右转向灯")
    hazard_lights_on: bool = Field(default=False, description="双闪")
    headlights_on: bool = Field(default=False, description="前灯")


class VehicleControlDTO(BaseModel):
    """车辆控制场景接口模型"""

    throttle: float = Field(default=0.0, description="油门（0~1）")
    brake: float = Field(default=0.0, description="刹车（0~1）")
    steering: float = Field(default=0.0, description="方向盘（-1~1）")
    gear: GearMode = Field(default=GearMode.DRIVE, description="档位")
    left_blinker_on: bool = Field(
        serialization_alias="Signal_Light_LeftBlinker",
        default=False,
        description="左转向灯",
    )
    right_blinker_on: bool = Field(
        serialization_alias="Signal_Light_RightBlinker",
        default=False,
        description="右转向灯",
    )
    hazard_lights_on: bool = Field(
        serialization_alias="Signal_Light_DoubleFlash",
        default=False,
        description="双闪",
    )
    headlights_on: bool = Field(
        serialization_alias="Signal_Light_FrontLight", default=False, description="前灯"
    )
    move_to_start: int = Field(serialization_alias="movetostart", description="重开")
    move_to_end: int = Field(serialization_alias="movetoend", description="跳关")


class SimCarMsgOutput(BaseModel):
    """code4 中使用的输出结构"""

    vehicle_control: VehicleControlDTO = Field(serialization_alias="VehicleControl")
    vla_extension: VLAExtensionOutput | None = Field(
        serialization_alias="VLAExtension",
        description="VLA 相关的输出，非 VLA 场景时为 None",
    )


class Code1(BaseModel):
    """code1 接口模型，接收静态信息"""

    code: Literal[1]
    map_info: MapConfig = Field(alias="MapInfo")
    vla_extension: VLAExtension | None = Field(
        default=None,
        alias="VLAExtension",
        description="VLA 扩展信息，如果为 None 则表示不是 VLA 场景",
    )


class Code2(BaseModel):
    """code2 接口模型，发送API已就绪"""

    code: Literal[2]


class Code3(BaseModel):
    """code3 接口模型，接收仿真动态信息"""

    code: Literal[3]
    sim_car_msg: SimCarMsg = Field(alias="SimCarMsg")


class Code4(BaseModel):
    """code4 接口模型，发送控制信息"""

    code: Literal[4]
    sim_car_msg: SimCarMsgOutput = Field(serialization_alias="SimCarMsg")


class Code5(BaseModel):
    """code5 接口模型，接收场景结束信息"""

    code: Literal[5]


@dataclass
class CameraFrame:
    """摄像头图像数据"""

    id: str  #: 对应 :attr:`CameraInfo.id`
    frame: np.ndarray  #: 图像数据



import logging
from pathlib import Path
from pydantic import TypeAdapter, Field
from typing import Annotated
from .sockets import ModelSocket, StreamingSocket, ConnectionClosedError
from .geometry import Vector3
from .models import (
    CameraFrame,
    SimCarMsgOutput,
    VehicleControl,
    VehicleControlDTO,
    RoadInfo,
    SceneStaticData,
    VLAExtensionOutput,
    Code1,
    Code2,
    Code3,
    Code4,
    Code5,
)

logger = logging.getLogger(__name__)


class SceneAPI:
    """SceneAPI 是与仿真环境通信的主要接口。

    该类封装了与仿真环境建立连接、获取场景信息、读取车辆状态以及发送控制命令等功能。
    使用流程通常是：创建实例 -> 连接 -> 获取静态数据 -> 进入主循环获取动态数据并发送控制命令。
    """

    def __init__(self):
        """初始化 SceneAPI 实例，但不会立即连接。
        需要调用 connect() 方法与仿真环境建立连接。
        """
        self._move_to_start = 0
        self._move_to_end = 0
        self._model_socket = ModelSocket("127.0.0.1", 5061)
        self._streaming_socket = StreamingSocket("127.0.0.1", 5063)

    def _load_static_data(self, code1: Code1):
        """读取文件内容，组装场景静态信息。

        从指定的地图配置中读取路径文件和地图文件，解析后组装成场景静态信息。

        :param code1: 场景发送的 code1 消息
        """
        map_info = code1.map_info
        dir_path = Path(map_info.path)
        route_path = dir_path / map_info.route
        with route_path.open("rb") as route_file:
            route = TypeAdapter(list[Vector3]).validate_json(route_file.read())
        map_path = dir_path / map_info.map
        with map_path.open("rb") as map_file:
            road_lines = TypeAdapter(list[RoadInfo]).validate_json(map_file.read())
        self._scene_static_data = SceneStaticData(
            route=route,
            roads=road_lines,
            sub_scenes=map_info.sub_scenes,
            vla_extension=code1.vla_extension,
        )

    def connect(self):
        """与场景建立连接，会产生阻塞，直到与场景连接成功。

        此方法会阻塞执行，直到成功与仿真环境建立连接并完成握手。
        连接成功后会加载场景静态数据，可通过 get_scene_static_data() 获取。
        """
        self._model_socket.accept()  # 连接 json socket
        self._streaming_socket.accept()  # 连接视频流
        code1: Code1 = self._model_socket.recv(Code1)
        self._load_static_data(code1)

    def get_scene_static_data(self):
        """获取场景静态信息，仅在 connect() 函数调用后可用

        此方法返回加载的场景静态数据，包括路线、道路信息和子场景信息。
        必须在调用 connect() 方法后才能使用。

        :return: 场景静态数据
        """
        return self._scene_static_data

    def main_loop(self):
        """生成器，每次迭代返回 :class:`~metacar.models.SimCarMsg` 和图像帧，场景结束时退出。

        此方法是一个生成器，每次迭代会返回当前的仿真车辆消息和摄像头图像帧。
        当场景结束或连接中断时，生成器会自动退出。

        :return: 元组 (sim_car_msg, frames)，其中:

            - sim_car_msg: :class:`~metacar.models.SimCarMsg` 对象，包含车辆状态、传感器数据等信息
            - frames: 当前相机视图的列表，每个元素为 :class:`~metacar.models.CameraFrame` 对象
        """
        # 先发送 code2，告知场景已经就绪
        self._model_socket.send(Code2(code=2), Code2)
        # 进入主循环，持续从场景接收消息
        try:
            while True:
                message: Code3 | Code5 = self._model_socket.recv(
                    Annotated[Code3 | Code5, Field(discriminator="code")]
                )
                if isinstance(message, Code5):
                    logger.info("场景结束")
                    return
                sim_car_msg = message.sim_car_msg
                frames = [
                    CameraFrame(id=camera_info.id, frame=self._streaming_socket.recv())
                    for camera_info in sim_car_msg.sensor.ego_rgb_cams
                ]
                yield sim_car_msg, frames
        except ConnectionClosedError:
            logger.warning("连接中断，退出场景")
            return
        finally:
            self._model_socket.close()
            self._streaming_socket.close()

    def set_vehicle_control(
        self, vc: VehicleControl, vla_extension: VLAExtensionOutput | None = None
    ):
        """发送车辆控制命令到仿真环境

        将给定的车辆控制命令发送到场景，用于控制车辆的油门、刹车、转向等行为。

        :param vc: 车辆控制命令，包含油门、刹车、转向等参数
        :param vla_extension: VLA 相关的输出，非 VLA 场景为 None
        """
        vc_dto = VehicleControlDTO(
            **vc.model_dump(),
            move_to_start=self._move_to_start,
            move_to_end=self._move_to_end,
        )
        sim_car_msg = SimCarMsgOutput(
            vehicle_control=vc_dto, vla_extension=vla_extension
        )
        self._model_socket.send(Code4(code=4, sim_car_msg=sim_car_msg), Code4)

    def retry_level(self):
        """重试关卡

        增加重试关卡计数器，在下一次发送控制命令时会通知场景重试当前关卡。
        """
        self._move_to_start += 1
        logger.info("重试关卡")

    def skip_level(self):
        """跳过关卡

        增加跳过关卡计数器，在下一次发送控制命令时会通知场景跳过当前关卡。
        """
        self._move_to_end += 1
        logger.info("跳过关卡")



import cv2
import numpy as np
import socket
import struct
import logging
from typing import Any
from pydantic import TypeAdapter

logger = logging.getLogger(__name__)


class ConnectionClosedError(ConnectionError):
    pass


class RawSocket:
    """
    一个简单的 TCP 服务器类，实现基于长度 + 内容格式的基本分包。
    长度为 4 字节的无符号整数，使用网络字节序（大端序）。
    仅支持单个客户端连接。
    """

    _HEADER_SIZE = 4  # 消息长度的 4 字节前缀

    def __init__(self, host: str, port: int):
        """
        初始化 socket 服务器。

        :param host: 服务器绑定的 IP 地址。
        :param port: 监听的端口号。
        """
        self._host = host
        self._port = port
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.bind((self._host, self._port))
        self._sock.listen()
        self._conn = None
        logger.info(f"监听 {self._host}:{self._port}")

    def accept(self):
        """
        接受一个新的客户端连接。
        如果已有客户端连接，则先关闭旧连接。
        """
        if self._conn:
            self._conn.close()
        self._conn, address = self._sock.accept()
        logger.info(f"{self._host}:{self._port}已连接到{address}")

    def send(self, data: bytes):
        """
        发送数据到客户端，数据前加上 4 字节的长度前缀。

        :param data: 需要发送的字节数据。
        """
        if not self._conn:
            raise ConnectionError("无客户端连接")
        length_prefix = struct.pack("!I", len(data))  # 将长度转换为 4 字节大端序
        self._conn.sendall(length_prefix + data)

    def recv(self) -> bytes:
        """
        从客户端接收数据，确保按照长度前缀读取完整的消息。

        该方法首先读取 4 字节的长度前缀，然后根据前缀指定的长度读取实际消息内容。
        如果连接已关闭，则返回空字节。如果没有客户端连接，则抛出 ConnectionError 异常。

        :return: 接收到的字节数据，如果连接已关闭则返回空字节。
        :raises ConnectionError: 当没有客户端连接时抛出。
        """
        if not self._conn:
            raise ConnectionError("无客户端连接")
        length_data = self._recv_exact(self._HEADER_SIZE)
        if not length_data:
            return b""
        message_length = struct.unpack("!I", length_data)[0]  # 解包 4 字节大端整数
        return self._recv_exact(message_length)

    def _recv_exact(self, size: int) -> bytes:
        """
        精确接收指定字节数的数据。

        :param size: 需要接收的字节数。
        :return: 接收到的字节数据，如果连接已关闭则返回空字节。
        """
        # recv 函数里面检查过了，这里简单 assert 一下，防止类型检查报错
        assert self._conn is not None, "无客户端连接"
        data = bytearray()
        while len(data) < size:
            chunk = self._conn.recv(size - len(data))
            if not chunk:
                return b""  # 连接已关闭
            data.extend(chunk)
        return bytes(data)

    def close(self):
        """
        关闭服务器 socket 及客户端连接。
        """
        if self._conn:
            self._conn.close()
        self._sock.close()
        logger.info(f"{self._host}:{self._port}已关闭")


class ModelSocket:
    """
    支持在每次收发时动态指定类型的 Pydantic Socket。

    发送时会将数据按照指定类型序列化为 JSON，
    接收时会解析为该类型的实例。
    """

    def __init__(self, host: str, port: int):
        self._raw_socket = RawSocket(host, port)

    def accept(self):
        return self._raw_socket.accept()

    def close(self):
        return self._raw_socket.close()

    def send(self, data: Any, type_: Any):
        """
        发送数据（自动 JSON 序列化）。

        :param data: 要发送的数据。
        :param type_: 数据的类型（可为 BaseModel、list[...]、tuple[...] 等）。
        """
        adapter = TypeAdapter(type_)
        json_bytes = adapter.dump_json(data, by_alias=True)
        self._raw_socket.send(json_bytes)

    def recv(self, type_: Any) -> Any:
        """
        接收数据（自动 JSON 反序列化）。

        :param type_: 目标类型（可为 BaseModel、list[...]、tuple[...] 等）。
        :return: 解析后的对象。
        :raises ConnectionClosedError: 当连接已关闭时抛出。
        """
        raw_data = self._raw_socket.recv()
        if not raw_data:
            raise ConnectionClosedError("连接已关闭")
        adapter = TypeAdapter(type_)
        return adapter.validate_json(raw_data)


class StreamingSocket:
    """
    支持单向接收视频流的 Socket。
    """

    def __init__(self, host: str, port: int):
        self._raw_socket = RawSocket(host, port)

    def accept(self):
        return self._raw_socket.accept()

    def close(self):
        return self._raw_socket.close()

    def recv(self) -> np.ndarray:
        """
        接收视频帧。

        :return: 接收到的视频帧。
        :rtype: numpy.ndarray
        :raises ConnectionClosedError: 当连接已关闭时抛出。
        """
        raw_image = self._raw_socket.recv()
        if not raw_image:
            raise ConnectionClosedError("连接已关闭")
        frame = cv2.imdecode(np.frombuffer(raw_image, np.uint8), cv2.IMREAD_COLOR)
        return frame



import logging
import keyboard
import math
from metacar import SceneAPI, GearMode, VehicleControl, Vector3, SimCarMsg

# 是否使用 GUI 界面
USE_GUI = True

if USE_GUI:
    from gui import Dashboard

logging.basicConfig(
    filename="autodrive.log",
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    level=logging.DEBUG,
    encoding="utf-8",
)
logger = logging.getLogger(__name__)

current_gear = GearMode.DRIVE
use_keyboard = True


def set_gear(gear: GearMode):
    global current_gear
    current_gear = gear


def toggle_keyboard():
    global use_keyboard
    use_keyboard = not use_keyboard


def get_vehicle_control_from_keyboard() -> VehicleControl:
    vc = VehicleControl()
    vc.gear = current_gear
    if keyboard.is_pressed("up") or keyboard.is_pressed("w"):
        value = 0.5 if keyboard.is_pressed("shift") else 1
        if current_gear == GearMode.DRIVE:
            vc.throttle = value
        elif current_gear == GearMode.REVERSE:
            vc.brake = value
    elif keyboard.is_pressed("down") or keyboard.is_pressed("s"):
        value = 0.5 if keyboard.is_pressed("shift") else 1
        if current_gear == GearMode.DRIVE:
            vc.brake = value
        elif current_gear == GearMode.REVERSE:
            vc.throttle = value
    if keyboard.is_pressed("left") or keyboard.is_pressed("a"):
        vc.steering = -1
    elif keyboard.is_pressed("right") or keyboard.is_pressed("d"):
        vc.steering = 1
    return vc


def calc_throttle_brake(
    current_speed: float, target_speed: float
) -> tuple[float, float]:
    K = 0.2
    B = 0.2
    acceleration = (target_speed - current_speed) * K + B
    if acceleration > 0:
        return min(acceleration, 1), 0
    return 0, min(-acceleration * 0.5, 1)


# Stanley 算法
def calc_steering(
    pos: Vector3, yaw: float, speed: float, trajectory: list[Vector3]
) -> float:
    K = 0.5
    for traj_pos in trajectory:
        if math.dist(traj_pos, pos) > K * speed:
            target_pos = traj_pos
            break
    else:
        target_pos = trajectory[-1]
    theta = (target_pos - pos).yaw_rad()
    steering_angle = (theta - yaw) % (2 * math.pi)
    if steering_angle > math.pi:
        steering_angle -= 2 * math.pi
    steering = math.degrees(-steering_angle) / 45
    return max(min(steering, 1), -1)


def get_vehicle_control_from_algorithm(msg: SimCarMsg) -> VehicleControl:
    vc = VehicleControl()
    vc.gear = GearMode.DRIVE
    # 目前只支持固定速度
    vc.throttle, vc.brake = calc_throttle_brake(msg.main_vehicle.speed, 15)
    vc.steering = calc_steering(
        Vector3(
            msg.pose_gnss.pos_x,
            msg.pose_gnss.pos_y,
            msg.pose_gnss.pos_z,
        ),
        -math.radians(msg.pose_gnss.ori_z),
        msg.main_vehicle.speed,
        msg.trajectory,
    )
    return vc


def main():
    api = SceneAPI()

    keyboard.add_hotkey("space", api.retry_level)
    keyboard.add_hotkey("n", api.skip_level)

    keyboard.add_hotkey("r", lambda: set_gear(GearMode.REVERSE))
    keyboard.add_hotkey("f", lambda: set_gear(GearMode.DRIVE))
    keyboard.add_hotkey("t", lambda: set_gear(GearMode.NEUTRAL))
    keyboard.add_hotkey("g", lambda: set_gear(GearMode.PARKING))

    keyboard.add_hotkey("c", toggle_keyboard)

    api.connect()
    static_data = api.get_scene_static_data()
    if USE_GUI:
        dashboard = Dashboard(static_data)
        logger.info("启动 GUI 界面")
    logger.info("开始场景")

    for sim_car_msg, frames in api.main_loop():
        if use_keyboard:
            vehicle_control = get_vehicle_control_from_keyboard()
        else:
            vehicle_control = get_vehicle_control_from_algorithm(sim_car_msg)
        api.set_vehicle_control(vehicle_control)
        if USE_GUI:
            dashboard.update(sim_car_msg)
    logger.info("结束场景")
    if USE_GUI:
        dashboard.quit()
        logger.info("关闭 GUI 界面")


if __name__ == "__main__":
    main()
